﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;

namespace NanoMvvm.Threading
{
    //-------------------------------------------------------------------------- 
    //  
    //  Copyright (c) Microsoft Corporation.  All rights reserved.  
    //  
    //  From the "Samples for Parallel Programming with the .NET Framework"
    //  Distributed under the MS-LPL license
    // 
    //--------------------------------------------------------------------------

    /// <summary>Provides a task scheduler that targets a specific SynchronizationContext.</summary> 
    public sealed class SynchronizationContextTaskScheduler : TaskScheduler
    {
        /// <summary>The queue of tasks to execute, maintained for debugging purposes.</summary> 
        private readonly ConcurrentQueue<Task> tasks;
        /// <summary>The target context under which to execute the queued tasks.</summary> 
        private readonly SynchronizationContext context;

        /// <summary>Initializes an instance of the SynchronizationContextTaskScheduler class.</summary> 
        public SynchronizationContextTaskScheduler() :
            this(SynchronizationContext.Current)
        {
        }

        /// <summary> 
        /// Initializes an instance of the SynchronizationContextTaskScheduler class 
        /// with the specified SynchronizationContext. 
        /// </summary> 
        /// <param name="context">The SynchronizationContext under which to execute tasks.</param> 
        public SynchronizationContextTaskScheduler(SynchronizationContext context)
        {
            if (context == null) throw new ArgumentNullException("context");
            this.context = context;
            tasks = new ConcurrentQueue<Task>();
        }

        /// <summary>Queues a task to the scheduler for execution on the I/O ThreadPool.</summary> 
        /// <param name="task">The Task to queue.</param> 
        protected override void QueueTask(Task task)
        {
            tasks.Enqueue(task);
            context.Post(delegate
            {
                Task nextTask;
                if (tasks.TryDequeue(out nextTask)) TryExecuteTask(nextTask);
            }, null);
        }

        /// <summary>Tries to execute a task on the current thread.</summary> 
        /// <param name="task">The task to be executed.</param> 
        /// <param name="taskWasPreviouslyQueued">Ignored.</param> 
        /// <returns>Whether the task could be executed.</returns> 
        protected override bool TryExecuteTaskInline(Task task, bool taskWasPreviouslyQueued)
        {
            return context == SynchronizationContext.Current && TryExecuteTask(task);
        }

        /// <summary>Gets an enumerable of tasks queued to the scheduler.</summary> 
        /// <returns>An enumerable of tasks queued to the scheduler.</returns> 
        protected override IEnumerable<Task> GetScheduledTasks()
        {
            return tasks.ToArray();
        }

        /// <summary>Gets the maximum concurrency level supported by this scheduler.</summary> 
        public override int MaximumConcurrencyLevel { get { return 1; } }
    }
}