﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;

namespace NanoMvvm.UnitTests
{
    [TestClass]
    public class DynamicViewModelTests
    {
        [TestMethod]
        public void BasicTest()
        {
            var dvm = new DynamicViewModel<TestClass>(new TestClass());

            dvm.WrapProperty((tc) => tc.PublicIntProperty);

            var descriptors = TypeDescriptor.GetProperties(dvm);

            var props = descriptors.Cast<PropertyDescriptor>().Select(prop => prop.Name).ToList();

            Assert.IsTrue(props.SequenceEqual(new[] { "WrappedModel", "PublicIntProperty" }));
        }

        [TestMethod]
        public void BasicGetSetWrappedValueTest()
        {
            var instance = new TestClass();

            var dvm = new DynamicViewModel<TestClass>(instance);

            dvm.WrapProperty((tc) => tc.PublicIntProperty);
            dvm.WrapProperty((tc) => tc.PublicStringProperty);

            var val = dvm.GetWrappedValue((tc) => tc.PublicIntProperty);

            Assert.AreEqual(val, 0);

            dvm.SetWrappedValue((tc) => tc.PublicIntProperty, 123);

            val = dvm.GetWrappedValue((tc) => tc.PublicIntProperty);

            Assert.AreEqual(val, 123);
            Assert.AreEqual(instance.PublicIntProperty, 123);
        }

        [TestMethod]
        public void BasicDependantPropertyNotifyTest()
        {
            var instance = new TestClass();

            var dvm = new DynamicViewModel<TestClass>(instance);

            dvm.WrapProperty((tc) => tc.PublicStringProperty, tc => tc.CompositeStringProperty);
            dvm.WrapProperty((tc) => tc.CompositeStringProperty);
            dvm.WrapProperty((tc) => tc.PublicIntProperty);

            dvm.PropertyChanged += instance.RecordPropertyChanged;

            dvm.SetWrappedValue((tc) => tc.PublicIntProperty, 123);

            Assert.IsTrue(instance.PropertiesRaised.SequenceEqual(new[] { "PublicIntProperty" }));

            dvm.SetWrappedValue((tc) => tc.PublicStringProperty, "data");

            Assert.IsTrue(instance.PropertiesRaised.SequenceEqual(new[] { "PublicIntProperty", "PublicStringProperty", "CompositeStringProperty" }));
        }

        [TestMethod]
        public void InheritedDynamicVMGetSetWrappedValueTest()
        {
            var instance = new TestClass();

            var vm = new InheritedDynamicTestClass(instance);

            vm.SetWrappedValue((tc) => tc.PublicIntProperty, 123);

            var val = vm.GetWrappedValue((tc) => tc.PublicIntProperty);

            Assert.AreEqual(val, 123);
        }

        [TestMethod]
        public void TypeDescriptorValueAccessTest()
        {
            var vm = new InheritedDynamicTestClass(new TestClass());

            var tdp = TypeDescriptor.GetProvider(vm);

            var td = tdp.GetTypeDescriptor(vm);

            var prop = td.GetProperties().Cast<PropertyDescriptor>().Where(pd => pd.Name == "PublicIntProperty").Single();

            var val = (int)prop.GetValue(vm);

            Assert.AreEqual(val, 0);

            prop.SetValue(vm, 123);

            val = (int)prop.GetValue(vm);

            Assert.AreEqual(val, 123);
        }

        private class TestClass
        {
            public TestClass()
            {
                PropertiesRaised = new List<string>();
            }

            public static string PublicStaticStringProperty { get; set; }

            public string PublicStringProperty { get; set; }

            public string CompositeStringProperty
            {
                get
                {
                    return PublicStringProperty + " + addition";
                }
            }

            public int PublicIntProperty { get; set; }

            public List<string> PropertiesRaised { get; set; }

            public void RecordPropertyChanged(object sender, PropertyChangedEventArgs e)
            {
                PropertiesRaised.Add(e.PropertyName);
            }
        }

        private class InheritedDynamicTestClass : DynamicViewModel<TestClass>
        {
            public InheritedDynamicTestClass(TestClass model)
                : base(model)
            {
                WrapProperty((tc) => tc.PublicStringProperty, tc => tc.CompositeStringProperty);
                WrapProperty((tc) => tc.CompositeStringProperty);
                WrapProperty((tc) => tc.PublicIntProperty);
            }
        }
    }
}