﻿using System.Collections.Generic;
using System.Diagnostics;
using NanoMvvm;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;

namespace NanoMvvm.UnitTests
{
    [TestClass]
    public class MediatorTest
    {
        private class TestClass
        {
            static TestClass()
            {
                Added = new List<int>();
            }

            public static IList<int> Added { get; private set; }

            public void TestActionMethod(int p)
            {
                Added.Add(p);
                Debug.WriteLine("TestActionMethod executed.");
            }

            public static void StaticTestActionMethod(int p)
            {
                Added.Add(p);
                Debug.WriteLine("StaticTestActionMethod executed.");
            }
        }

        [TestMethod]
        [DeploymentItem("NanoMvvm.dll")]
        public void RegisterHandlerInstanceTest()
        {
            Mediator.Default.Clear();

            var tc = new TestClass();

            Assert.IsTrue(Mediator.Default.RegisterHandler<int>(tc.TestActionMethod));

            //already added
            Assert.IsFalse(Mediator.Default.RegisterHandler<int>(tc.TestActionMethod));
        }

        [TestMethod]
        public void UnregisterHandlerInstanceTest()
        {
            Mediator.Default.Clear();

            var tc = new TestClass();

            for (int i = 0; i < 10; i++)
            {
                Assert.IsTrue(Mediator.Default.RegisterHandler<int>(tc.TestActionMethod));

                //already added
                Assert.IsFalse(Mediator.Default.RegisterHandler<int>(tc.TestActionMethod));


                Assert.IsTrue(Mediator.Default.UnregisterHandler<int>(tc.TestActionMethod));

                //already removed
                Assert.IsFalse(Mediator.Default.UnregisterHandler<int>(tc.TestActionMethod));
            }
        }

        [TestMethod]
        public void UnregisterHandlerStaticTest()
        {
            Mediator.Default.Clear();

            for (int i = 0; i < 10; i++)
            {
                Assert.IsTrue(Mediator.Default.RegisterHandler<int>(TestClass.StaticTestActionMethod));

                //already added
                Assert.IsFalse(Mediator.Default.RegisterHandler<int>(TestClass.StaticTestActionMethod));


                Assert.IsTrue(Mediator.Default.UnregisterHandler<int>(TestClass.StaticTestActionMethod));

                //already removed
                Assert.IsFalse(Mediator.Default.UnregisterHandler<int>(TestClass.StaticTestActionMethod));
            }
        }

        [TestMethod]
        public void NotifyInstanceTest()
        {
            Mediator.Default.Clear();

            TestClass.Added.Clear();

            var tc = new TestClass();

            Mediator.Default.RegisterHandler<int>(tc.TestActionMethod);
            Mediator.Default.RegisterHandler<int>(TestClass.StaticTestActionMethod);

            var count = Mediator.Default.NotifySubscribers(123);

            Assert.AreEqual(TestClass.Added.Count, 2);

            Assert.AreEqual(TestClass.Added[0], 123);
            Assert.AreEqual(TestClass.Added[1], 123);

            Assert.AreEqual(count, 2);
        }

        [TestMethod]
        public void NotifyStaticTest()
        {
            Mediator.Default.Clear();

            TestClass.Added.Clear();

            Mediator.Default.RegisterHandler<int>(TestClass.StaticTestActionMethod);

            Mediator.Default.NotifySubscribers(123);

            Assert.IsTrue(TestClass.Added.Count == 1);

            Assert.IsTrue(TestClass.Added[0] == 123);
        }
    }
}
