﻿using System;
using System.Collections.Generic;
using System.Linq;

using Microsoft.VisualStudio.TestTools.UnitTesting;

using NanoMvvm.Collections;

namespace NanoMvvm.UnitTests
{
    internal static class ListHelper
    {
        private static readonly Random Rng = new Random();

        public static void Shuffle<T>(this IList<T> list)
        {
            int n = list.Count;
            while (n > 1)
            {
                n--;
                int k = Rng.Next(n + 1);
                T value = list[k];
                list[k] = list[n];
                list[n] = value;
            }
        }
    }

    [TestClass]
    public class SortedKeyListTests
    {
        [TestMethod]
        public void BasicTest()
        {
            var collection = new SortedKeyedList<int, TestClass>(tc => tc.Key);

            var list = Enumerable.Range(0, 5000).Select(i => new TestClass(i)).ToList();

            list.Shuffle();

            foreach (var testClass in list)
            {
                collection.Add(testClass);
            }

            var result = collection.SelectInterval(200, 300).ToList();

            CollectionAssert.AreEqual(result, collection.Where(c => c.Key >= 200 && c.Key <= 300).OrderBy(c => c.Key).ToArray());
        }

        [TestMethod]
        public void BasicIntervalTest()
        {
            var now = DateTime.Now;

            var collection = new SortedKeyedList<DateTime, TestClass2>(tc => tc.Key);

            var list = Enumerable.Range(0, 5000).Select(i => new TestClass2(now + TimeSpan.FromHours(i))).ToList();

            list.Shuffle();

            foreach (var testClass in list)
            {
                collection.Add(testClass);
            }

            var result = collection.SelectInterval(now, now + TimeSpan.FromHours(100)).ToList();

            CollectionAssert.AreEqual(result, collection.Where(c => c.Key >= now && c.Key <= now + TimeSpan.FromHours(100)).OrderBy(c => c.Key).ToArray());
        }

        [TestMethod]
        public void AdvancedIntervalTest()
        {
            var now = DateTime.Now;

            var collection = new SortedKeyedList<DateTime, TestClass2>(tc => tc.Key);

            var list = Enumerable.Range(0, 5000).Select(i => new TestClass2(now + TimeSpan.FromHours(i))).ToList();

            list.Shuffle();

            foreach (var testClass in list)
            {
                collection.Add(testClass);
            }

            var result0 = collection;
            var collection0 = collection.ToArray();

            var start = now;
            var result1 = collection.SelectInterval(start, start + TimeSpan.FromHours(100));
            var collection1 = collection.Where(c => c.Key >= start && c.Key <= start + TimeSpan.FromHours(100)).OrderBy(c => c.Key).ToArray();

            start = start.AddHours(100);
            var result2 = collection.SelectInterval(start, start + TimeSpan.FromHours(100));
            var collection2 = collection.Where(c => c.Key >= start && c.Key <= start + TimeSpan.FromHours(100)).OrderBy(c => c.Key).ToArray();

            start = start.AddHours(100);
            var result3 = collection.SelectInterval(start, start + TimeSpan.FromHours(100));
            var collection3 = collection.Where(c => c.Key >= start && c.Key <= start + TimeSpan.FromHours(100)).OrderBy(c => c.Key).ToArray();

            result0.Any();
            result1.Any();
            result2.Any();
            result3.Any();

            CollectionAssert.AreEqual(result1.ToArray(), collection1);
            CollectionAssert.AreEqual(result2.ToArray(), collection2);
            CollectionAssert.AreEqual(result3.ToArray(), collection3);

            CollectionAssert.AreEqual(result0.ToArray(), collection0);
        }



        public class TestClass
        {
            public TestClass(int key)
            {
                Key = key;
            }

            public int Key { get; set; }

            public override string ToString()
            {
                return Key.ToString();
            }
        }

        public class TestClass2
        {
            public TestClass2(DateTime key)
            {
                Key = key;
            }

            public DateTime Key { get; set; }

            public override string ToString()
            {
                return Key.ToString();
            }
        }
    }
}