﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace NanoMvvm.UnitTests
{
    [TestClass]
    public class WeakEventListenerTest
    {
        private class EventTestClass
        {
            public event EventHandler TestEvent;

            public void RaiseEvent()
            {
                TestEvent(null, null);
            }

            public int EventSubscribersCount()
            {
                return TestEvent == null ? 0 : TestEvent.GetInvocationList().Count();
            }

            internal static readonly List<string> Messages = new List<string>();

            internal const string InsideEventMsg = "Inside event!";
            internal const string DeregisteringMsg = "Deregistered handler wrapper";
        }

        private class ListenerClass
        {
            public void DoStuff(object o, EventArgs e)
            {
                EventTestClass.Messages.Add(EventTestClass.InsideEventMsg);
            }
        }

        [TestMethod]
        public void WrapWeakEventListenerCollectedTest()
        {
            EventTestClass.Messages.Clear();

            var tc = new EventTestClass();
            var lc = new ListenerClass();

            tc.TestEvent += WeakEventListener.Wrap(lc.DoStuff,
                delegate(EventHandler e)
                {
                    tc.TestEvent -= e;
                    EventTestClass.Messages.Add(EventTestClass.DeregisteringMsg);
                });

            Assert.AreEqual(tc.EventSubscribersCount(), 1);

            tc.RaiseEvent();

            lc = null;

            GC.Collect();

            tc.RaiseEvent();

            Assert.IsTrue(EventTestClass.Messages.SequenceEqual(new[] { EventTestClass.InsideEventMsg, EventTestClass.DeregisteringMsg }));

            Assert.AreEqual(tc.EventSubscribersCount(), 0);
        }

        [TestMethod]
        public void WeakSourceEventListenerCollectedTest()
        {
            var tc = new EventTestClass();
            var lc = new ListenerClass();

            var wr = new WeakReference<EventTestClass>(tc);

            tc.TestEvent += WeakEventListener.Wrap(lc.DoStuff,
                delegate(EventHandler e)
                {
                    tc.TestEvent -= e;
                    EventTestClass.Messages.Add(EventTestClass.DeregisteringMsg);
                });

            Assert.IsTrue(wr.IsTargetAlive);

            tc = null;

            GC.Collect();

            Assert.IsFalse(wr.IsTargetAlive);
        }
    }
}